import re

from textual.app import ComposeResult
from textual.widgets import Label
from textual.widgets import Static

from .leds import OnOffLed


class AEBState(Static):
    """A widget to monitor the state of the AEBs."""

    def compose(self) -> ComposeResult:
        yield Label()
        yield Label("ON/OFF", classes="title")
        yield Label("INIT", classes="title")
        yield Label("POWER-UP", classes="title")
        yield Label("POWER-DOWN", classes="title")
        yield Label("CONFIG", classes="title")
        yield Label("IMAGE", classes="title")
        yield Label("PATTERN", classes="title")

        yield Label("AEB1")
        yield OnOffLed(id="aeb1_onoff")
        yield OnOffLed(id="aeb1_init")
        yield OnOffLed(id="aeb1_power_up")
        yield OnOffLed(id="aeb1_power_down")
        yield OnOffLed(id="aeb1_config")
        yield OnOffLed(id="aeb1_image")
        yield OnOffLed(id="aeb1_pattern")

        yield Label("AEB2")
        yield OnOffLed(id="aeb2_onoff")
        yield OnOffLed(id="aeb2_init")
        yield OnOffLed(id="aeb2_power_up")
        yield OnOffLed(id="aeb2_power_down")
        yield OnOffLed(id="aeb2_config")
        yield OnOffLed(id="aeb2_image")
        yield OnOffLed(id="aeb2_pattern")

        yield Label("AEB3")
        yield OnOffLed(id="aeb3_onoff")
        yield OnOffLed(id="aeb3_init")
        yield OnOffLed(id="aeb3_power_up")
        yield OnOffLed(id="aeb3_power_down")
        yield OnOffLed(id="aeb3_config")
        yield OnOffLed(id="aeb3_image")
        yield OnOffLed(id="aeb3_pattern")

        yield Label("AEB4")
        yield OnOffLed(id="aeb4_onoff")
        yield OnOffLed(id="aeb4_init")
        yield OnOffLed(id="aeb4_power_up")
        yield OnOffLed(id="aeb4_power_down")
        yield OnOffLed(id="aeb4_config")
        yield OnOffLed(id="aeb4_image")
        yield OnOffLed(id="aeb4_pattern")

    def set_state(self, aeb_state_type: str, aeb_state: bool):

        if (aeb_nr := get_aeb_nr(aeb_state_type)) is None:
            self.notify(f"Couldn't derive AEB unit number from {aeb_state_type=}", severity="warning")
            return

        old_onoff_state = self.query_one(f"#aeb{aeb_nr}_onoff", OnOffLed).state
        old_other_state = self.query_one(f"#{aeb_state_type}", OnOffLed).state

        # When the state is init, config, image, or pattern, set the onoff led also, otherwise
        # onoff will be cleared below.

        if not aeb_state_type.endswith("_onoff"):
            self.query_one(f"#aeb{aeb_nr}_onoff", OnOffLed).state = True
            if old_onoff_state is False:
                self.notify(f"AEB{aeb_nr} is Powered ON")

        # Clear the current states except ONOFF

        self.query_one(f"#aeb{aeb_nr}_init", OnOffLed).state = False
        self.query_one(f"#aeb{aeb_nr}_power_up", OnOffLed).state = False
        self.query_one(f"#aeb{aeb_nr}_power_down", OnOffLed).state = False
        self.query_one(f"#aeb{aeb_nr}_config", OnOffLed).state = False
        self.query_one(f"#aeb{aeb_nr}_image", OnOffLed).state = False
        self.query_one(f"#aeb{aeb_nr}_pattern", OnOffLed).state = False

        self.query_one(f"#{aeb_state_type}", OnOffLed).state = aeb_state


def get_aeb_nr(string: str):
    """
    Returns the first digit that matched in the given string. Intended to match the AEB number.

    Returns None when no match.
    """
    # The first digit that is matched in 'string' will be returned.
    #
    # This is used for:
    # - button.id -> 'btn-aeb[1-4]-[\w-]+'
    # -

    match = re.search(r'\d', string)
    if match:
        aeb_nr = int(match.group())
    else:
        aeb_nr = None

    return aeb_nr



from __future__ import annotations

import logging
import pickle
import queue
import threading
import time
import traceback
from queue import Queue
from typing import TYPE_CHECKING

import zmq
from egse.dpu.fdpu import FastCameraDPUProxy
from egse.fee.ffee import HousekeepingData
from egse.fee.ffee import aeb_state
from egse.fee.ffee import f_fee_mode
from egse.reg import RegisterMap
from egse.settings import Settings
from egse.setup import load_setup
from egse.zmq import MessageIdentifier

from .messages import AebStateChanged
from .messages import DebModeChanged
from .messages import DtcInModChanged
from .messages import ExceptionCaught
from .messages import TimeoutReached

if TYPE_CHECKING:
    from .app import FastFEEApp

_LOGGER = logging.getLogger("egse.f-fee-tui")

dpu = Settings.load("DPU Processor")


class Command(threading.Thread):
    def __init__(self, app: 'FastFEEApp', command_q: Queue) -> None:
        super().__init__()
        self._app = app
        self._command_q = command_q
        self._f_dpu: FastCameraDPUProxy | None = None
        self._canceled = threading.Event()

    def run(self) -> None:

        with FastCameraDPUProxy() as self._f_dpu:
            while True:
                if self._canceled.is_set():
                    break
                try:
                    target, command, args, kwargs = self._command_q.get_nowait()
                    self.execute_command(target, command, args, kwargs)
                    self._command_q.task_done()
                except queue.Empty:
                    time.sleep(0.1)  # Sleep briefly to avoid busy-waiting
                    continue
                except Exception as exc:
                    _LOGGER.error(f"Caught and exception: {exc}")
                    tb = traceback.format_exc()
                    self._app.post_message(ExceptionCaught(exc, tb))

    def cancel(self) -> None:
        self._canceled.set()

    def execute_command(self, target: str, command: str, args: list, kwargs: dict):
        if target == "DPU":
            try:
                return getattr(self._f_dpu, command)(*args, **kwargs)
            except AttributeError as exc:
                _LOGGER.error(f"No such command for DPU: {command}", exc_info=True)


class Monitor(threading.Thread):
    def __init__(self, app: 'FastFEEApp') -> None:
        self._app = app
        self._canceled = threading.Event()
        self.hostname = dpu.HOSTNAME
        self.port = dpu.DATA_DISTRIBUTION_PORT

        self.previous_deb_mode = f_fee_mode.ON_MODE
        self.previous_aeb_state = {}

        super().__init__()

    def run(self) -> None:

        n_timeouts = 0
        """The timeout is 1s, we count the number of timeouts to detect if the DPU or N-FEE might be dead or stalled."""

        context = zmq.Context.instance()
        receiver = context.socket(zmq.SUB)
        receiver.setsockopt_string(zmq.SUBSCRIBE, "")
        receiver.connect(f"tcp://{self.hostname}:{self.port}")

        setup = load_setup()

        while True:
            if self._canceled.is_set():
                break

            socket_list, _, _ = zmq.select([receiver], [], [], timeout=1.0)

            if receiver in socket_list:
                n_timeouts = 0
                try:
                    sync_id, pickle_string = receiver.recv_multipart()
                    sync_id = int.from_bytes(sync_id, byteorder='big')
                    data = pickle.loads(pickle_string)
                    self.handle_messages(sync_id, data, setup)
                except Exception as exc:
                    tb = traceback.format_exc()
                    self._app.post_message(ExceptionCaught(exc, tb))

            if len(socket_list) == 0:
                n_timeouts += 1
                if n_timeouts > 3:  # at least a timecode should arrive every 2.5s
                    self._app.post_message(TimeoutReached("Timeout reached after 3s on monitoring channel."))
                    n_timeouts = 0

        receiver.disconnect(f"tcp://{self.hostname}:{self.port}")
        receiver.close()

    def cancel(self) -> None:
        self._canceled.set()

    def handle_messages(self, sync_id, data, setup):

        if sync_id == MessageIdentifier.F_FEE_REGISTER_MAP:

            # How can we be sure the Register Map is properly synchronised with the F-FEE?

            register_map, _ = data
            register_map = RegisterMap("F-FEE", memory_map=register_map, setup=setup)

            t0 = register_map["DEB_DTC_IN_MOD_2", "T0_IN_MOD"]
            t1 = register_map["DEB_DTC_IN_MOD_2", "T1_IN_MOD"]
            t2 = register_map["DEB_DTC_IN_MOD_2", "T2_IN_MOD"]
            t3 = register_map["DEB_DTC_IN_MOD_2", "T3_IN_MOD"]
            t4 = register_map["DEB_DTC_IN_MOD_1", "T4_IN_MOD"]
            t5 = register_map["DEB_DTC_IN_MOD_1", "T5_IN_MOD"]
            t6 = register_map["DEB_DTC_IN_MOD_1", "T6_IN_MOD"]
            t7 = register_map["DEB_DTC_IN_MOD_1", "T7_IN_MOD"]

            self._app.log(f"{t0=}, {t1=}, {t2=}, {t3=}, {t4=}, {t5=}, {t6=}, {t7=}")

            self._app.post_message(DtcInModChanged(t0, t1, t2, t3, t4, t5, t6, t7))

        elif sync_id == MessageIdentifier.SYNC_HK_DATA:

            cmd, aeb_id, data, timestamp = data

            if cmd == 'command_deb_read_hk':
                hk_data = HousekeepingData("DEB", data, setup)
                deb_mode = f_fee_mode(hk_data["STATUS", "OPER_MOD"])
                self._app.log(f"DEB_MODE = {deb_mode.name}")
                self._app.post_message(DebModeChanged(deb_mode))
            elif cmd == 'command_aeb_read_hk':
                aeb_id = aeb_id[0]  # this comes from the args, so it's a list
                hk_data = HousekeepingData(aeb_id, data, setup)
                aeb_status = hk_data["STATUS", "AEB_STATUS"]
                self._app.log(f"AEB_STATE = {aeb_state(aeb_status).name}")

                # Power in handled by the DEB_DTC_AEB_ONOFF state from the DEB register map above.

                if aeb_status == aeb_state.OFF:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_onoff", False))
                elif aeb_status == aeb_state.INIT:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_init", True))
                elif aeb_status == aeb_state.POWER_UP:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_power_up", True))
                elif aeb_status == aeb_state.POWER_DOWN:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_power_down", True))
                elif aeb_status == aeb_state.CONFIG:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_config", True))
                elif aeb_status == aeb_state.IMAGE:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_image", True))
                elif aeb_status == aeb_state.PATTERN:
                    self._app.post_message(AebStateChanged(f"{aeb_id.lower()}_pattern", True))



from textual import events
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Label

from f_fee_tui.services import services


class InfoBar(Horizontal):

    DEFAULT_CSS = """
    InfoBar {
        height: 1;
        dock: bottom;
        Label {
            margin: 0 1 0 0; 
        }
    }
    """

    def compose(self) -> ComposeResult:
        yield Label("cm_cs", id="cm_cs")
        yield Label("🔴", id="cm_cs_active", classes="status")
        yield Label("sm_cs", id="sm_cs")
        yield Label("🔴", id="sm_cs_active", classes="status")
        yield Label("pm_cs", id="pm_cs")
        yield Label("🔴", id="pm_cs_active", classes="status")
        yield Label("syn_cs", id="syn_cs")
        yield Label("🔴", id="syn_cs_active", classes="status")
        yield Label("dump", id="data_dump")
        yield Label("🔴", id="data_dump_active", classes="status")
        yield Label("dpu_cs", id="dpu_cs")
        yield Label("🔴", id="dpu_cs_active", classes="status")
        yield Label("data", id="data")
        yield Label("🔴", id="data_active", classes="status")

    def on_mount(self, event: events.Mount) -> None:

        for name, props in services.items():
            lbl = self.query_one(f"#{name}", Label)
            tooltip = props.get('description', '')
            self.app.log(f"Added tooltip to ")
            lbl.tooltip = tooltip

    def set_active(self, name: str, is_active: bool):
        self.query_one(f"#{name}_active", Label).update('🔴' if is_active else '🟢')



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Footer, Header, Tree
from textual.widgets import Label

CHARACTERS = {
    "Paul": "Son of Leto Atreides",
    "Jessica": "Mother of Paul Atreides",
    "Chani": "Daughter of Liet-Kynes",
}


class MyTree(Tree):
    def watch_hover_line(self, previous: int, line: int) -> None:
        super().watch_hover_line(previous, line)
        try:
            self.tooltip = self.get_node_at_line(line).data
        except KeyError:
            self.tooltip = None


class TreeScreen(Screen):
    CSS = """
        TreeScreen {
            Tooltip {
                visibility: hidden;
            }

            &.show_tooltips Tooltip {
                visibility: visible;
            }
        }
    """
    BINDINGS = [("t", "toggle_tooltips", "Toggle Tooltips")]

    DEFAULT_CLASSES = "show_tooltips"

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Label("What is this?", id="tooltip-label")
        yield MyTree("Dune")

    def on_mount(self) -> None:
        tree = self.query_one(MyTree)
        tree.root.expand()
        characters = tree.root.add("Characters", expand=True)
        for name, description in CHARACTERS.items():
            characters.add_leaf(name, data=description)

        self.query_one("#tooltip-label", Label).tooltip = "This is a Label with a tooltip"

    def action_toggle_tooltips(self) -> None:
        self.toggle_class("show_tooltips")
        if self.has_class("show_tooltips"):
            self.notify("Showing tooltips")
        else:
            self.notify("Tooltips disabled")


class TreeApp(App):
    def on_mount(self) -> None:
        self.push_screen(TreeScreen())


if __name__ == "__main__":
    app = TreeApp()
    app.run()



import asyncio
import itertools

from textual import work
from textual.app import App
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Footer
from textual.widgets import Header
from textual.widgets import Label
from textual.widgets import Markdown


class MyApp(App):
    """A Textual app to monitor and command the PLATO F-FEE."""

    DEFAULT_CSS = """
    Label {
        dock: bottom;
    }
    """

    BINDINGS = [
        Binding(key="f1", action="help", description="Help", show=True, priority=True),
    ]

    def __init__(self):
        super().__init__()
        self._poll_timer = None

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()
        with Vertical():
            yield Label("Hello, World!", id="lbl-hello")

    def on_mount(self):
        self._poll_timer = self.set_timer(1.0, self.poll_services)

    def action_help(self) -> None:
        self.app.push_screen(HelpScreen())

    @work()
    async def poll_services(self):
        names = itertools.cycle(["John", "Bart", "melissa", "Kris"])

        while True:
            await asyncio.sleep(2.0)
            self.query_one("#lbl-hello", Label).update(f"Hello, {next(names)}!")


HELP = """
# A Simple Help Modal Screen.

This is an example App tp serve as a MRE.
Use ESCAPE to dismiss.
"""


class HelpScreen(ModalScreen[None]):
    """Modal dialog that shows the application's help."""

    DEFAULT_CSS = """
    HelpScreen {
        align: center middle;
    }
    HelpScreen > Vertical {
        border: thick grey 50%;
        width: 60%;
        height: 80%;
        background: $boost;
    }

    """

    BINDINGS = [Binding("escape", "dismiss", "", show=False)]

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Markdown(HELP)

    def action_dismiss(self, result=None):
        self.dismiss(result)


if __name__ == '__main__':
    app = MyApp()
    app.run()



from textual import events
from textual.app import ComposeResult
from textual.widgets import Button
from textual.widgets import Static

# Tooltips

BTN_DEB_ON = """
Press this button to bring the F-FEE in ON mode. 

ON mode can only be reached from STANDBY mode or FULL-IMAGE PATTERN mode.
"""
BTN_DEB_STANDBY = """
Press this button to bring the F-FEE in STANDBY mode.

STANDBY mode can only be reached from ON mode, and FULL-IMAGE mode.
"""
BTN_DEB_FULL_IMAGE = """
Press this button to bring the F-FEE into FULL-IMAGE mode.

FULL-IMAGE mode can only be reached from STANDBY mode.
"""
BTN_DEB_FULL_IMAGE_PATTERN = """
Press this button to bring the F-FEE into FULL-IMAGE PATTERN mode.

FULL-IMAGE PATTERN mode can only be reached from ON mode.
"""
BTN_DEB_IMMEDIATE_ON = """
Press this button to bring the F-FEE immediately into ON mode.

The DEB will return to ON mode immediately, regardless of the mode it is in. 
The AEBs will return to INIT state and power down the CCDs and the VASP.

Then, the AEBs will be Powered OFF.
"""

class DEBCommand(Static):

    HELP = """Press the button to change the DEB mode."""

    def compose(self) -> ComposeResult:
        yield Button("ON", id='btn-deb-on', classes='command')
        yield Button("FULL IMAGE PATTERN", id='btn-deb-full-image-pattern', classes='command')
        yield Button("STANDBY", id='btn-deb-standby', classes='command')
        yield Button("FULL IMAGE", id='btn-deb-full-image', classes='command')
        yield Button("IMMEDIATE ON", id='btn-deb-immediate-on', classes='command')

    def _on_mount(self, event: events.Mount) -> None:
        self.query_one("#btn-deb-on").tooltip = BTN_DEB_ON
        self.query_one("#btn-deb-standby").tooltip = BTN_DEB_STANDBY
        self.query_one("#btn-deb-full-image").tooltip = BTN_DEB_FULL_IMAGE
        self.query_one("#btn-deb-full-image-pattern").tooltip = BTN_DEB_FULL_IMAGE_PATTERN
        self.query_one("#btn-deb-immediate-on").tooltip = BTN_DEB_IMMEDIATE_ON
